#!/usr/bin/env python

from math import pi
import numpy as np
import robot_calc_functions as rcf
import matplotlib.pyplot as plt


UR5thstart =[0.1,0.1,0.1,0.1,0.1,0.1]
UR5thend =[pi/2,pi/2,pi/2,pi/2,pi/2,pi/2]
T = 2
N = 101
tscalemethod = 3

#Cubic time scaled joint trajectory
JointTraj = rcf.JointTrajectory(UR5thstart, UR5thend, T, N, tscalemethod)

#*********Plotting the result************************
jointlist = []
for i in range(0,len(JointTraj)):
    jointlist.append(JointTraj[i][0])

timestamp = np.linspace(0,2,N-1)
plt.plot(timestamp, jointlist, label = "Cubic Time Scaling")
plt.ylim (0,2)
plt.legend(loc = 'upper left')
plt.xlabel("Time")
plt.ylabel("S(t)")
plt.title("Plot of S(t) for Cubic Time Scaling")
plt.show()
#**************************************************

#Quintic time scaled joint trajectory
tscalemethod = 5

JointTraj = rcf.JointTrajectory(UR5thstart, UR5thend, T, N, tscalemethod)

#*******Plotting the results**********************
jointlist = []
for i in range(0,len(JointTraj)):
    jointlist.append(JointTraj[i][0])

timestamp = np.linspace(0,2,N-1)
plt.plot(timestamp, jointlist, label = "Quintic Time Scaling")

plt.ylim (0,2)
plt.legend(loc = 'upper left')
plt.xlabel("Time")
plt.ylabel("S(t)")
plt.title("Plot of S(t) for Quintic Time Scaling")
plt.show()
#**************************************************


#Storing in a csv file for animation purposes
f = open('quest5_ur5joints.csv', 'w')
f.write ("time,shoulder_pan_joint,shoulder_lift_joint,elbow_joint,wrist_1_joint,wrist_2_joint,wrist_3_joint\n")
t = 0

for i in range (N-1):
    f.write(str(t))
    t+=0.02
    f.write(",")
    f.write(str(JointTraj[i][0]))
    f.write(",")
    f.write(str(JointTraj[i][1]))
    f.write(",")
    f.write(str(JointTraj[i][2]))
    f.write(",")
    f.write(str(JointTraj[i][3]))
    f.write(",")
    f.write(str(JointTraj[i][4]))
    f.write(",")
    f.write(str(JointTraj[i][5]))
    f.write("\n")
f.close()





