#!/usr/bin/env python

from math import pi
import numpy as np
import robot_calc_functions as rcf
import matplotlib.pyplot as plt

UR5M = [[1,0,0,-0.817],
      [0,0,-1,-0.191],
      [0,1,0,-0.006],
     [0,0,0,1]]

UR5Sb = [[0,1,0,0.191,0,0.817],[0,0,1,0.095,-0.817,0],[0,0,1,0.095,-0.392,0],[0,0,1,0.095,0,0],[0,-1,0,-0.082,0,0],[0,0,1,0,0,0]]

UR5thstart =[0.101,0.101,0.101,0.101,0.101,0.101]
UR5thend =[pi/2,pi/2,pi/2,pi/2,pi/2,pi/2]

#Find Xstart and Xend using FK
Xstart = rcf.FKinBody(UR5M, UR5Sb, UR5thstart)
Xend = rcf.FKinBody(UR5M, UR5Sb, UR5thend)

#Printing answers to the terminal
print "Xstart rounded to 3 decimal places:"
print np.round(Xstart,3)
print
print "Xend rounded to 3 decimal places:"
print np.round(Xend,3)

T = 2
N = 101
tscalemethod = 5

#Find the Cartesian Trajectories
Xst = rcf.CartesianTrajectory(Xstart, Xend, T, N, tscalemethod)

JointAngles=[]

UR5Ew =0.00174533
UR5Ev =0.0005
UR5maxiterations = 15

#Get all the joint angles from the last IK solution in each iteration
for i in range (N-1):
    jointangles,iterations = rcf.IKinBody(UR5Sb, UR5M, Xst[i], UR5thstart, UR5Ew, UR5Ev, UR5maxiterations)    
    JointAngles.append(jointangles[iterations-1])
    UR5thstart = jointangles[iterations-1]

#Write to file for animation
f = open('Quest6_ur5joints.csv', 'w')
f.write ("time,shoulder_pan_joint,shoulder_lift_joint,elbow_joint,wrist_1_joint,wrist_2_joint,wrist_3_joint\n")
t = 0

for i in range (N-1):
    f.write(str(t))
    t+=0.02
    f.write(",")
    f.write(str(JointAngles[i][0]))
    f.write(",")
    f.write(str(JointAngles[i][1]))
    f.write(",")
    f.write(str(JointAngles[i][2]))
    f.write(",")
    f.write(str(JointAngles[i][3]))
    f.write(",")
    f.write(str(JointAngles[i][4]))
    f.write(",")
    f.write(str(JointAngles[i][5]))
    f.write("\n")
f.close()



#PLOTTING
for j in range (0,6):
    jointlist = []
    for i in range(0,len(JointAngles)):
        jointlist.append(JointAngles[i][j])
    timestamp = np.linspace(0,2,N-1)
    plt.plot(timestamp, jointlist, label = ("Joint %d" %(j+1)))
plt.ylim (-6,8)
plt.legend(loc = 'upper left')
plt.xlabel("Time")
plt.ylabel("S(t)")
plt.title("Plot of S(t) for 6 Joint Angles of the UR5")
plt.show()


