#!/usr/bin/python
import numpy as np
from math import cos, acos, sin, tan, pi, sqrt

def matmult(*x):#Returns the dot product of the inputs
    '''
Example Input:
([1,2,3], [3,2,1])
Output:
10
    '''
    return reduce(np.dot, x)



def Magnitude(V):#Takes in a vector and returns its length
    '''
Example Input: 
V = [1,2,3]
Output:
3.74165738677
    '''
    try:
        length = 0
        for i in range(len(V)):
            length += (V[i]**2)
        return ((length)**0.5)
    except:
        print ("Input is not a vector")



def Normalise(V):#Takes in a vector and scales it to a unit vector
    '''
Example Input: 
V = [1,2,3]
Output:
[0.2672612419124244, 0.5345224838248488, 0.8017837257372732]
    '''
    try:
        length = Magnitude(V)
        if (length == 0.):
            for i in range(len(V)):
                V[i] = 0.0
            return V
        for i in range(len(V)):
            V[i] = V[i]/length
        return V
    except:
        print ("Input is not a vector")



def Det_3(R):#Takes a 3x3 matrix and returns the determinant
    '''
Example Input: 
R = [[0, 0,1],
    [1, 0, 0],
    [0, 1, 0]]
Output:
1.0
    '''
    try:
        return np.linalg.det(R)
    except:
        print("Matrix is not square")




def RotInv(R):#Takes a 3x3 rotation matrix and returns the transpose (inverse)
    '''
Example Input: 
R = [[0, 0,1],
    [1, 0, 0],
    [0, 1, 0]]
Output:
[[0, 1, 0], 
[0, 0, 1],
[1, 0, 0]]
    '''
    #Test for determinant = 1 +- 0.05 error
    if (Det_3(R)*1.0>=0.95 and Det_3(R)*1.0<=1.05):
        Rt = [[R[0][0],R[1][0],R[2][0]],[R[0][1],R[1][1],R[2][1]],[R[0][2],R[1][2],R[2][2]]]
    if(matmult(R,Rt) == np.eye(3)).all:#Test to make sure R*Rt=I 
        return Rt




def VecToso3(V):#Takes a vector(angular velocity) and returns the skew sym matrix in so3
    '''
Example Input: 
V = [1,2,3]
Output:
[[0, -3, 2],
 [3, 0, -1],
 [-2, 1, 0]]
    '''
    try:
        if (len(V)==3):
            return [[0,-V[2],V[1]], [V[2],0,-V[0]], [-V[1],V[0],0]]
        else:
            print ("This vector is the wrong size")
    except:
        print ("This vector cannot be converted to a skew sym matrix")



def so3ToVec(R):#Takes a skew sym matrix in so3 and returns the corresponding vector (angular velo)
    '''
Example Input: 
R = [[0, -3, 2],
 [3, 0, -1],
 [-2, 1, 0]]
Output:
[1, 2, 3]
    '''
    try:
        if (R[0][0]==0 and R[1][1]==0 and R[2][2]==0 and
        R[0][1]==-R[1][0] and R[2][0]==-R[0][2] and R[1][2]==-R[2][1]):#Check for skew sym matrix
                return [R[2][1],R[0][2],R[1][0]]
        else:
            print ("This is not a skew symetric matrix")
    except:
        print ("Input martix is the wrong shape")


def AxisAng3(V):#Takes 3vector exponential coords and returns unit rotation axis w and rotation amount theta
    '''
Example Input: 
V = [1,2,3]
Output:
([0.2672612419124244, 0.5345224838248488, 0.8017837257372732], -->unit rotation axis w
 3.7416573867739413) -->theta
    '''
    if (len(V)==3):
        theta = Magnitude(V)
        if (theta == 0):
            return ([0,0,0],0)
        else:
            return (Normalise(V), theta)#(w, theta)
    else:
        print ("This vector is the wrong size")



    
def MatrixExp3(V):#Takes 3vector of exp coords and returns R' matrix achieved by rotation (I initial)
    #Rodriguez R = I + sin(theta)*w + (1-cos(theta))*w^2
    '''
Example Input: 
p = [0*0.524,0.866*0.524,0.5*0.524]
Output:
[[ 0.86583049 -0.25017423  0.43330176],
[ 0.25017423  0.96645615  0.05809795],
[-0.43330176  0.05809795  0.89937434]]
    '''
    if (len(V)==3):
        w,theta = AxisAng3(V)
        return np.eye(3) + matmult(VecToso3(w),np.sin(theta)) + matmult(VecToso3(w),VecToso3(w))*(1-np.cos(theta))        
    else:
        print ("This vector is the wrong size")




def MatrixLog3(R):#Takes R matrix and returns r (where r = w*theta)
    '''
Example Input: 
R = [[0, 0,1],[1, 0, 0],[0, 1, 0]]

Output:
[1.2091995761561456, 1.2091995761561456, 1.2091995761561456]
    '''
    if (len(R)==3):
        try:
            Rtrace = R[0][0]+R[1][1]+R[2][2]
            theta = np.arccos((Rtrace - 1)/2.0)
            w = 1/(2*np.sin(theta))*np.array([R[2][1]-R[1][2], R[0][2]-R[2][0], R[1][0]-R[0][1]])
            if any(map(np.isinf, w)) or any(map(np.isnan, w)):
                theta = 0
                w = 3*[1/np.sqrt(3)]
            return [w[0]*theta,w[1]*theta,w[2]*theta]
        except:
            print("Matrix cannot be converted")        
    else:
        print ("This matrix is the wrong size")





def RpToTrans (R,p):#Takes rotation matrix and position p and returns transformation matrix T SE(3)
    '''
Example Input: 
R = [[1, 0, 0], [0, 0, -1], [0, 1, 0]]
p = [1,2,5]
Output:
[[1, 0, 0, 1],
 [0, 0, -1, 2],
 [0, 1, 0, 5],
 [0, 0, 0, 1]]
    '''
    try:
        return [ [R[0][0],R[0][1],R[0][2],p[0]] , [R[1][0],R[1][1],R[1][2],p[1]], [R[2][0],R[2][1],R[2][2],p[2]] , [0,0,0,1]]
    except:
        print ("Input rotation matrix or position vector are the wrong size")




def TransToRp (T):#Take transformation matrix T SE(3) and returns R rotation matrix and p position vector
    '''
Example Input: 
T = [[1,0,0,0],
     [0,0,-1,0],
     [0,1,0,3],
     [0,0,0,1]]
Output:
([[1, 0, 0], -->R
 [0, 0, -1], -->R
 [0, 1, 0]], -->R 
[0, 0, 3]) -->P
    '''
    try:
        return ([ [T[0][0],T[0][1],T[0][2]],
                  [T[1][0],T[1][1],T[1][2]],
                  [T[2][0],T[2][1],T[2][2]]],
                [T[0][3],T[1][3],T[2][3]])
    except:
        print ("Input Transformation matrix is the wrong size")




def TransInv(T):#Takes an element of SE3 and returns its inverse
    '''
Example Input: 
T = [[1,0,0,0],
     [0,0,-1,0],
     [0,1,0,3],
     [0,0,0,1]]
Output:
[[1, 0, 0, 0],
 [0, 0, 1, -3],
 [0, -1, 0, 0],
 [0, 0, 0, 1]]
    '''
    try:
        R,p = TransToRp(T)
        Rt = RotInv(R)
        pt = np.dot(Rt, p)
        return [ [Rt[0][0],Rt[0][1],Rt[0][2],-pt[0]],
                 [Rt[1][0],Rt[1][1],Rt[1][2],-pt[1]],
                 [Rt[2][0],Rt[2][1],Rt[2][2],-pt[2]],
                 [0,0,0,1]]
    except:
        print ("Input matrix is not an element of SE3")





def VecTose3(V):#Takes a 6vector (Spacial Velo) and returns 4x4 matrix element of se3
    '''
Example Input: 
V = [1,2,3,4,5,6]
Output:
[[0, -3, 2, 4], [3, 0, -1, 5], [-2, 1, 0, 6], [0, 0, 0, 0]]
    '''
    if (len(V)==6):
        try:
           R = VecToso3([V[0],V[1],V[2]])
           return [ [R[0][0],R[0][1],R[0][2],V[3]] , [R[1][0],R[1][1],R[1][2],V[4]], [R[2][0],R[2][1],R[2][2],V[5]] , [0,0,0,0]]
        except:
            print("Input vector is not a spatial velocity representation")        
    else:
        print ("Input vector is the wrong size")






def se3ToVec(Sp):#Takes a 4x4 matrix element of se3 and returns 6vector (Spacial Velo)
    '''
Example Input: 
Sp =[[-1,0,0,0],  [0,1,0,6],  [0,0,-1,2],  [0,0,0,1]]
Output:
[0, 0, 0, 0, 6, 2]
    '''
    try:
        return ([Sp[2][1],Sp[0][2],Sp[1][0],
                 Sp[0][3],Sp[1][3],Sp[2][3]])
    except:
        print ("Input is not an element of se3")





def Adjoint(T):#Takes a T element of SE3 and returns 6x6 adjoint representation
    '''
Example Input: 
T = [[1,0,0,0], [0,0,-1,0], [0,1,0,3], [0,0,0,1]]
Output:
[[1, 0, 0, 0, 0, 0],
[0, 0, -1, 0, 0, 0],
[0, 1, 0, 0, 0, 0],
[0, 3, 0, 1, 0, 0],
[0, 0, 0, 0, 0, -1],
[-3, 0, 0, 0, 1, 0]]
    '''
    try:
        R,p = TransToRp(T)
        pp = VecToso3(p)
        Rt = np.dot(pp,R) 
        return [[R[0][0],R[0][1],R[0][2],0,0,0],
                [R[1][0],R[1][1],R[1][2],0,0,0],
                 [R[2][0],R[2][1],R[2][2],0,0,0],
                  [Rt[0][0],Rt[0][1],Rt[0][2],R[0][0],R[0][1],R[0][2]],
                  [Rt[1][0],Rt[1][1],Rt[1][2],R[1][0],R[1][1],R[1][2]],
                  [Rt[2][0],Rt[2][1],Rt[2][2],R[2][0],R[2][1],R[2][2]]]
    except:
        print ("Input T is not an element of SE3")




def ScrewToAxis(q,s,h):#Takes a point q, a unit axis s, and screw pitch h and returns corresponding screw axis
    '''
Example Input: 
q = [3,0,0]
s = [0,0,1]
h = 2
Output:
[[0], [0], [1], [0], [-3], [2]]
    '''
    try:
        sq=np.cross(s,q)
        return [[s[0]],
                [s[1]],
                [s[2]],
                [(-sq[0]+h*s[0])],
                [(-sq[1]+h*s[1])],
                [(-sq[2]+h*s[2])]]
    except:
        print ("Inputs are the wrong size. -->  qE3, sE3, h scaler")





def AxisAng6(V):#Takes 6vector Stheta and returns screw axis s and distance travelled theta
    '''
Example Input: 
V = [1,0,0,1,2,3]
Output:
([1.0, 0.0, 0.0, 1.0, 2.0, 3.0], 1.0) --> First the srew axis and then the theta
    '''
    if (len(V)==6):
        theta = Magnitude([V[0],V[1],V[2]])
        if (theta == 0):
            theta = Magnitude([V[3],V[4],V[5]])
            if (theta==0):
                return ([0,0,0,0,0,0],0)
            return ([0,0,0,V[3]/theta*1.0,V[4]/theta*1.0,V[5]/theta*1.0],
                    theta)#(S,theta)
        else:
            return ([V[0]/theta*1.0,V[1]/theta*1.0,V[2]/theta*1.0,
                     V[3]/theta*1.0,V[4]/theta*1.0,V[5]/theta*1.0],
                    theta)#(S,theta)
    else:
        print ("This vector is the wrong size")




def MatrixExp6(T):#Takes a 6vector Stheta and returns a T matrix SE(3) 
    #Rodriguez R = I + sin(theta)*w + (1-cos(theta))*w^2
    '''
Example Input: 
Stheta = [0,(1/(2**0.5)),(1/(2**0.5)),1,2,3]
Output:
[[0.54030230586813965, -0.59500983952938591, 0.59500983952938591, 1.1665263416243543],
[0.59500983952938591, 0.77015115293406988, 0.22984884706593017, 2.4043198644125101],
[-0.59500983952938591, 0.22984884706593017, 0.77015115293406988, 2.5956801355874903],
[0, 0, 0, 1]]
    '''
    s,th = AxisAng6(T)
    w = [s[0],s[1],s[2]]
    UL = np.eye(3) + matmult(VecToso3(w),np.sin(th)) + matmult(VecToso3(w),VecToso3(w))*(1-np.cos(th))
    UR = matmult(np.eye(3),th) + matmult(VecToso3(w),(1-np.cos(th))) + matmult(VecToso3(w),VecToso3(w))*(th-np.sin(th))
    UR = np.dot(UR,[s[3],s[4],s[5]])

    return [[UL[0][0],UL[0][1],UL[0][2],UR[0]],
            [UL[1][0],UL[1][1],UL[1][2],UR[1]],
             [UL[2][0],UL[2][1],UL[2][2],UR[2]],
              [0,0,0,1]]


def MatrixLog6(T):#Takes a T matrix SE(3) and returns a 6vector Stheta
    '''
Example Input: 
T = [[1,0,0,0], [0,0,-1,0], [0,1,0,3], [0,0,0,1]]
Output:
[1.5707963267948966, 0.0, 0.0, 0.0, 2.3561944901923448, 2.3561944901923457]
    '''
    R,p = TransToRp(T)
    Rtrace = R[0][0]+R[1][1]+R[2][2]
    if(R==np.eye(3)).all():
        w=0
        v=Normalise(p)
        th=Magnitude(p)
    
    elif(Rtrace == -1):
        th = pi
        w = MatrixLog3(R)
        G = (1/th)*np.eye(3) - 0.5*np.asarray(VecToso3(w)) + ((1/th)-((1/(tan(th/2.0)))/2.0))*(matmult(VecToso3(w),VecToso3(w)))
        v = np.dot(G,p)

    else:
        th = acos((Rtrace-1)/2.0)
        w = so3ToVec((1/(2*np.sin(th)))*(np.subtract(R, RotInv(R))))
        G = (1/th)*np.eye(3) - 0.5*np.asarray(VecToso3(w)) + ((1/th)-((1/(tan(th/2.0)))/2.0))*(matmult(VecToso3(w),VecToso3(w)))
        v = np.dot(G,p)     

    return ([w[0]*th,w[1]*th,w[2]*th,v[0]*th,v[1]*th,v[2]*th])


def FKinFixed(M, S, th):#Takes M (position, orientation of end effector frame at home position), List of Screw axes S, List of joint coords theta and returns T representing the end effector frame with specified joint angles (ITO Fixed Frame)
    '''
Example Input: 
M = [[-1,0,0,0], [0,1,0,6], [0,0,-1,2], [0,0,0,1]]
S = [[0,0,1,4,0,0],[0,0,0,0,1,0],[0,0,-1,-6,0,-0.1]]
th =[(pi/2.0),3,pi]
Output:
[[ -1.14423775e-17   1.00000000e+00   0.00000000e+00  -5.00000000e+00],
[  1.00000000e+00   1.14423775e-17   0.00000000e+00   4.00000000e+00],
[ 0.          0.         -1.          1.68584073],
[ 0.  0.  0.  1.]]
    '''
    T = np.eye(4)
    for i in range(len(S)):
        T = np.dot(T,MatrixExp6(np.asarray(S[i])*th[i]))
        
    T = np.dot(T,M)
    return T



def FKinBody(M, S, th):#Takes M (position, orientation of end effector frame at home position), List of Screw axes S, List of joint coords theta and returns T representing the end effector frame with specified joint angles (ITO Body Frame)
    '''
Example Input: 
M = [[-1,0,0,0], [0,1,0,6], [0,0,-1,2], [0,0,0,1]]
Sb = [[0,0,-1,2,0,0],[0,0,0,0,1,0],[0,0,1,0,0,0.1]]
th =[(pi/2.0),3,pi]
Output:
[[ -1.14423775e-17   1.00000000e+00   0.00000000e+00  -5.00000000e+00],
[  1.00000000e+00   1.14423775e-17   0.00000000e+00   4.00000000e+00],
[ 0.          0.         -1.          1.68584073],
[ 0.  0.  0.  1.]]
    '''
    T = np.eye(4)
    T = np.dot(M,T)
    for i in range(len(S)):
        T = np.dot(T,MatrixExp6(np.asarray(S[i])*th[i]))       
    
    return T


def FixedJacobian(S, th):#Takes a set of screw axes, "S", and joint angles, "theta" for the robot in fixed space frame and returns the Fixed Jacobian (element of 6xn real numbers)
    '''
Example Input: 
S = [[0,0,1,0,0.2,0.2],[1,0,0,2,0,3,],[0,1,0,0,2,1],[1,0,0,0.2,0.3,0.4]]
th = [0.2,1.1,0.1,1.2]
Output:
[[ 0.          0.98006658 -0.09011564  0.95749426]
 [ 0.          0.19866933  0.4445544   0.28487557]
 [ 1.          0.          0.89120736 -0.04528405]
 [ 0.          1.95218638 -2.21635216 -0.51161537]
 [ 0.2         0.43654132 -2.43712573  2.77535713]
 [ 0.2         2.96026613  3.23573065  2.22512443]]
    '''
    if (len(S)==len(th)):
        Jac = [[0]*6 for i in range(len(S))]
        Jac[0] = S[0]
        for i in range(1,len(S)):
            if (len(S[i])==6):               
                JacAdj = MatrixExp6(matmult(S[0],th[0]))
                for j in range (i-1):                 
                    JacAdj = matmult(JacAdj, MatrixExp6(matmult(S[j+1],th[j+1]))) 
                JacAdj = Adjoint(JacAdj)
                Jac[i] = matmult(JacAdj,S[i]) 
            else:
                print "There is the wrong number of elements in the inputted screw axes"
                return
        return np.asarray(Jac).T
    else:
        print "The input has a different number of screw axes and joint angles"
        return
    



def BodyJacobian(Sb, th):#Takes a set of screw axes, "Sb", and joint angles, "theta" for the robot in body frame and returns the body Jacobian (element of 6xn real numbers)
    '''
Example Input: 
Sb = [[0,0,1,0,0.2,0.2],[1,0,0,2,0,3,],[0,1,0,0,2,1],[1,0,0,0.2,0.3,0.4]]
th = [0.2,1.1,0.1,1.2]
Output:
[[-0.04528405  0.99500417  0.          1.        ]
 [ 0.74359313  0.09304865  0.36235775  0.        ]
 [-0.66709716  0.03617541 -0.93203909  0.        ]
 [ 2.32586047  1.66809     0.56410831  0.2       ]
 [-1.44321167  2.94561275  1.43306521  0.3       ]
 [-2.06639565  1.82881722 -1.58868628  0.4       ]]
    '''
    if (len(Sb)==len(th)):
        Jac = [[0]*6 for i in range(len(Sb))]
        for i in range(len(Sb)):
            if (len(Sb[i])==6):
                JacAdj = MatrixExp6(matmult(matmult(Sb[len(Sb)-1],-1),th[len(Sb)-1])) 
                for j in range (len(Sb)-2-i):
                    JacAdj = matmult(JacAdj,MatrixExp6(matmult(matmult(Sb[len(Sb)-j-2],-1),th[len(Sb)-j-2]))) 
               
                JacAdj = Adjoint(JacAdj)
                Jac[i] = matmult(JacAdj,Sb[i])  
            else:
                print "There is the wrong number of elements in the inputted screw axes"
                return
        return np.asarray(Jac).T
    else:
        print "The input has a different number of screw axes and joint angles"
        return
    

def IKinBody(Sb, M, Tsd, th0, Ew, Ev, maxiterations):#Newton-Raphson method
#Takes a set of screw axes, "Sb" (expressed in body frame)
#The end-efffector zero config M element of SE(3)
#Desired end-effector config, Tsd
#Set of initial guesses for joint angles, th0
#Small Scale values, Ew, and Ev
#And maximum number of iterations before terminating the algoritm, maxiterations
#Returns joint angle matrix stepping from initial guess to closest final
#Also returns number of iterations it took to complete
    '''
Example Input: 
Sb = [[0,1,0,0.191,0,0.817],[0,0,1,0.095,-0.817,0],[0,0,1,0.095,-0.392,0],[0,0,1,0.095,0,0],[0,-1,0,-0.082,0,0],[0,0,1,0,0,0]]
M = [[1,0,0,-0.817],
    [0,0,-1,-0.191],
    [0,1,0,-0.006],
    [0,0,0,1]]
Tsd = [[0,1,0,-0.6], [0,0,-1,0.1], [-1,0,0,0.1], [0,0,0,1]]
th0 =[0,0,0,0,0,0]
Ew = 0.01
Ev = 0.001
maxiterations = 15
Output:
Joint Angles:
[[ 0.          0.          0.          0.          0.          0.        ]
 [-0.35618115 -0.53509301  0.46808006  1.3933597  -0.35618115 -2.89714308]
 [-0.39868394 -1.00512251  1.67641395 -0.43382608 -0.10004692 -1.80083873]
 [-0.51596739 -1.06243563  1.73068825 -1.63011762 -0.50190576 -0.60743995]
 [-0.49320033 -0.92317355  1.50775676 -0.73033224 -0.28927461 -1.42027077]
 [-0.47224864 -0.81826198  1.36519422 -0.45481125 -0.46719691 -1.66238721]
 [-0.46921905 -0.83447622  1.39525223 -0.56107486 -0.46731326 -1.57056352]]
Number of iterations used = 7
    '''

    thf =[]#Final return matrix
    thf.append(th0)
    Vb = MatrixLog6(matmult(TransInv(FKinBody(M, Sb, th0)),Tsd))
    wb  = Magnitude ([Vb[0],Vb[1],Vb[2]])
    vb = Magnitude ([Vb[3],Vb[4],Vb[5]])
    for i in range (maxiterations):
        if (wb>Ew and vb>Ev):
            th0 = np.add(th0, matmult(np.linalg.pinv(BodyJacobian(Sb, th0)),Vb))
            thf.append(th0)
            Vb = MatrixLog6(matmult(TransInv(FKinBody(M, Sb, th0)),Tsd))
            wb  = Magnitude ([Vb[0],Vb[1],Vb[2]])
            vb = Magnitude ([Vb[3],Vb[4],Vb[5]])
        else:
            return (np.asarray(thf),i+1)
    return (np.asarray(thf),i+1)


def IKinFixed(S, M, Tsd, th0, Ew, Ev, maxiterations):
#Takes a set of screw axes, "S" (expressed in the fixed frame)
#The base zero config M element of SE(3)
#Desired end-effector config, Tsd
#Set of initial guesses for joint angles, th0
#Small Scale values, Ew, and Ev
#And maximum number of iterations before terminating the algoritm, maxiterations
#Returns joint angle matrix stepping from initial guess to closest final
#Also returns number of iterations it took to complete
    '''
Example Input: 
S = [[0,0,1,0,0,0],[0,1,0,0,0,0],[0,0,1,0,0,0],[0,1,0,-0.550,0,0.045],[0,0,1,0,0,0],[0,1,0,-0.850,0,0],[0,0,1,0,0,0]]
M = [[1,0,0,0],
    [0,1,0,0],
    [0,0,1,0.910],
    [0,0,0,1]]
Tsd = [[1,0,0,0.8], [0,1,0,0.8], [0,0,1,0.8], [0,0,0,1]]
th0 = [0.1,0.1,0.1,0.1,0.1,0.1,0.1]
Ew = 0.01
Ev = 0.001
maxiterations = 15
Output:
[[  0.1          0.1          0.1          0.1          0.1            0.1         0.1       ]
 [ 12.21217639   2.41372705 -13.19153794  -4.45402839  -9.35597001    1.83678262  10.53885263]
 [ 14.87459909  -0.77112272 -10.50490467  -3.35354561  -8.28598086    1.20056452   9.96138821]
 [ 25.04740378   2.66827226 -14.68339383 -11.30425771  -1.29541269   -4.1117507   -2.70764655]
 [ 29.85754414   3.29470753 -13.70090952  -7.66858993   0.63933284   -7.94215937  -1.19573046]
 [ 25.50071017  -0.56443382 -26.2463175  -10.10335778   3.80566171   -8.17272735  -8.4753596 ]
 [ 25.39092704   1.97129861 -24.19527556  -9.79326874   6.7535951    -9.18660627  -8.59725828]
 [ 24.77157483  -4.67378231 -29.96159974 -14.4707883    2.04669587    1.41425108  -8.32432601]
 [ 27.28270026  -6.17830591 -31.14606319 -14.5511706   -0.14100282    1.63904949  -9.48530385]
 [ 29.35967957  -6.76040861 -27.20264781 -13.44212453   1.62801482    0.54211303 -14.24626656]
 [ 31.02215197  -6.70239646 -26.78412844 -11.52764633   3.26145454    2.34846143 -19.92852956]]
Number of iterations used = 10
    '''
    thf =[]#Final return matrix
    thf.append(th0)
    Vb = MatrixLog6(matmult(TransInv(FKinFixed(M, S, th0)),Tsd))
    wb  = Magnitude ([Vb[0],Vb[1],Vb[2]])
    vb = Magnitude ([Vb[3],Vb[4],Vb[5]])
    for i in range (maxiterations):
        if (wb>Ew and vb>Ev):
            bodyJac = matmult(Adjoint(TransInv(FKinFixed(M, S, th0))),FixedJacobian(S, th0))
            th0 = np.add(th0, matmult(np.linalg.pinv(bodyJac),Vb))
            thf.append(th0)
            Vb = MatrixLog6(matmult(TransInv(FKinFixed(M, S, th0)),Tsd))
            wb  = Magnitude ([Vb[0],Vb[1],Vb[2]])
            vb = Magnitude ([Vb[3],Vb[4],Vb[5]])
        else:
            return (np.asarray(thf),i+1)
    return (np.asarray(thf),i+1)


def CubicTimeScaling(T, t):#Takes total time T and current time t and returns path parameter s (Zero velocity begin and end) at time t
    '''
Example Input: 
T = 2
t = 0.6
Output:
-0.216
    '''
    #a0 and a1 are zero
    #Setting a2 and a3
    a2 = (3/(T**2))
    a3 = (-2/(T**3))
    return ((a2*(t**2))+(a3*(t**3)))


def QuinticTimeScaling(T, t):#Takes total time T and current time t and returns path parameter s (Zero velocity and zero acceleration begin and end) at time t
    '''
Example Input: 
T = 2
t = 0.6
Output:
0.0864
    '''
    #a0, a1, and a2 are zero
    #Setting a3, a4, and a5
    a3 = (10/(T**3))
    a4 = (-15/(T**4))
    a5 = (6/(T**5))
    return ((a3*(t**3))+(a4*(t**4))+(a5*(t**5)))


def JointTrajectory(thstart, thend, T, N, tscalemethod):#Takes start and end joint variable thstart, thend, the time of the motion T, the number of points N in discrete representation (must be greater than 1, start and end), and the time scaling method (cubic or quintic (3,5)). Returns a trajectory matrix (N rows) with each row is an n-vector with of joint angles. (Time between each row is T/N-1).
    '''
Example Input: 
thstart = [1,0,0,1,1,0.2,0,1]
thend = [1.2,0.5,0.6,1.1,2,2,0.9,1]
T = 4
N = 6
tscalemethod = 3
Output rounded to 3 decimal places:
[[ 1.     0.     0.     1.     1.     0.2    0.     1.   ]
 [ 1.021  0.052  0.062  1.01   1.104  0.387  0.094  1.   ]
 [ 1.07   0.176  0.211  1.035  1.352  0.834  0.317  1.   ]
 [ 1.13   0.324  0.389  1.065  1.648  1.366  0.583  1.   ]
 [ 1.179  0.448  0.538  1.09   1.896  1.813  0.806  1.   ]
 [ 1.2    0.5    0.6    1.1    2.     2.     0.9    1.   ]]
    '''
    #Check required method
    if (tscalemethod != 3 and tscalemethod != 5):
            print "Please input either 3 or 5 for the time scaling method"
            return
    #Check that N has enough points
    if (N < 2):
            print "Please input a bigger N"
            return
    #Check that thestart and thend are the same size
    if (len(thstart)!=len(thend)):
            print "Please enter a theta start the same size as your theta end"
            return

    #Convert to floats
    T += 0.0
    N += 0.0

    timegap = T/(N-1)#difference in time between each point
    joint_angle_matrix=[]#empty array to store joint angles
    
    curr_time=0
    while(curr_time<=T):
        if(tscalemethod==3):#Cubic
            joint_angle_matrix.append(np.add(matmult(1-CubicTimeScaling(T, curr_time), thstart),(matmult(CubicTimeScaling(T, curr_time),thend))))
        else:#Quintic
            joint_angle_matrix.append(np.add(matmult(1-QuinticTimeScaling(T, curr_time), thstart),(matmult(QuinticTimeScaling(T, curr_time),thend))))
        curr_time += timegap
            
    return np.asarray(joint_angle_matrix)


def ScrewTrajectory(Xstart, Xend, T, N, tscalemethod):#Takes start and end end effector configurationXstart and Xend, the time of the motion T, the number of points N in discrete representation (must be greater than 1, start and end), and the time scaling method (cubic or quintic (3,5)). Returns a list of N matrices in SE3 representing the trajectories from start to finish (Time between each row is T/N-1).
    '''
Example Input: 
Xstart = [[1,0,0,1],[0,1,0,0],[0,0,1,1],[0,0,0,1]]
Xend = [[0,0,1,0.1],[1,0,0,0],[0,1,0,4.1],[0,0,0,1]]
T = 5
N = 4
tscalemethod = 3
Output:
[[[ 1.     0.     0.     1.   ]
  [ 0.     1.     0.     0.   ]
  [ 0.     0.     1.     1.   ]
  [ 0.     0.     0.     1.   ]]

 [[ 0.904 -0.25   0.346  0.441]
  [ 0.346  0.904 -0.25   0.529]
  [-0.25   0.346  0.904  1.601]
  [ 0.     0.     0.     1.   ]]

 [[ 0.346 -0.25   0.904 -0.117]
  [ 0.904  0.346 -0.25   0.473]
  [-0.25   0.904  0.346  3.274]
  [ 0.     0.     0.     1.   ]]

 [[-0.     0.     1.     0.1  ]
  [ 1.    -0.     0.    -0.   ]
  [ 0.     1.    -0.     4.1  ]
  [ 0.     0.     0.     1.   ]]]
    '''
    #Check required method
    if (tscalemethod != 3 and tscalemethod != 5):
            print "Please input either 3 or 5 for the time scaling method"
            return
    #Check that N has enough points
    if (N < 2):
            print "Please input a bigger N"
            return
    #Check that Xstart and Xend are the same size
    if (len(Xstart)!=len(Xend)):
            print "Please enter a Xstart the same size as your Xend"
            return

    #Convert to floats
    T += 0.0
    N += 0.0

    timegap = T/(N-1)#difference in time between each point
    trjectory_matrix=[]#empty array to store trajectory

    curr_time=0
    while(curr_time<=T):
        if(tscalemethod==3):#Cubic
            trjectory_matrix.append(matmult(Xstart,MatrixExp6(matmult(MatrixLog6(matmult(TransInv(Xstart),Xend)),CubicTimeScaling(T, curr_time)))))
        else:#Quintic
            trjectory_matrix.append(matmult(Xstart,MatrixExp6(matmult(MatrixLog6(matmult(TransInv(Xstart),Xend)),QuinticTimeScaling(T, curr_time)))))
        curr_time += timegap
            
    return np.asarray(trjectory_matrix)



def CartesianTrajectory(Xstart, Xend, T, N, tscalemethod):#Takes start and end end effector configurationXstart and Xend, the time of the motion T, the number of points N in discrete representation (must be greater than 1, start and end), and the time scaling method (cubic or quintic (3,5)). Returns a list of N matrices in SE3 representing the trajectories from start to finish but with the end effector following a straight line decoupled from the rotational motion (Time between each row is T/N-1).
    '''
Example Input: 
Xstart = [[1,0,0,1],[0,1,0,0],[0,0,1,1],[0,0,0,1]]
Xend = [[0,0,1,0.1],[1,0,0,0],[0,1,0,4.1],[0,0,0,1]]
T = 5
N = 4
tscalemethod = 5
Output rounded to 3 decimal places:
[[[ 1.     0.     0.     1.   ]
  [ 0.     1.     0.     0.   ]
  [ 0.     0.     1.     1.   ]
  [ 0.     0.     0.     1.   ]]

 [[ 0.937 -0.214  0.277  0.811]
  [ 0.277  0.937 -0.214  0.   ]
  [-0.214  0.277  0.937  1.651]
  [ 0.     0.     0.     1.   ]]

 [[ 0.277 -0.214  0.937  0.289]
  [ 0.937  0.277 -0.214  0.   ]
  [-0.214  0.937  0.277  3.449]
  [ 0.     0.     0.     1.   ]]

 [[-0.     0.     1.     0.1  ]
  [ 1.    -0.     0.     0.   ]
  [ 0.     1.    -0.     4.1  ]
  [ 0.     0.     0.     1.   ]]]
    '''
    #Check required method
    if (tscalemethod != 3 and tscalemethod != 5):
            print "Please input either 3 or 5 for the time scaling method"
            return
    #Check that N has enough points
    if (N < 2):
            print "Please input a bigger N"
            return
    #Check that Xstart and Xend are the same size
    if (len(Xstart)!=len(Xend)):
            print "Please enter a Xstart the same size as your Xend"
            return

    #Convert to floats
    T += 0.0
    N += 0.0

    timegap = T/(N-1)#difference in time between each point
    trjectory_matrix=[]#empty array to store trajectory

    #Separate the Rotational motion and the straight line
    Rstart, Pstart = TransToRp(Xstart)
    Rend, Pend = TransToRp(Xend)

    curr_time=0
    while(curr_time<=T):
        if(tscalemethod==3):#Cubic
            Pcurrent = np.add(matmult((1-CubicTimeScaling(T, curr_time)),Pstart), (matmult(CubicTimeScaling(T,curr_time),Pend)))
            Rcurrent = (matmult(Rstart,MatrixExp3(matmult(MatrixLog3(matmult(RotInv(Rstart),Rend)),CubicTimeScaling(T, curr_time)))))            
        else:#Quintic
            Pcurrent = np.add(matmult(1-QuinticTimeScaling(T, curr_time), Pstart),(matmult(QuinticTimeScaling(T, curr_time),Pend)))
            Rcurrent = (matmult(Rstart,MatrixExp3(matmult(MatrixLog3(matmult(RotInv(Rstart),Rend)),QuinticTimeScaling(T, curr_time)))))
        
        #Put R and P back together and add it to the matrix
        trjectory_matrix.append(RpToTrans(Rcurrent, Pcurrent))
        curr_time += timegap
            
    return np.asarray(trjectory_matrix)
